#include "string.h"
#include "gtkcsdwindow.h"

struct _GtkCsdWindowPrivate
{
};

G_DEFINE_TYPE (GtkCsdWindow, gtk_csd_window, GTK_TYPE_WINDOW);

static void
paint_background_and_frame (GtkCsdWindow *window,
                            cairo_t      *cr)
{
  GtkWidget *widget = GTK_WIDGET (window);
  GtkStyleContext *context;
  gint width, height;

  width = gtk_widget_get_allocated_width (widget);
  height = gtk_widget_get_allocated_height (widget);
  context = gtk_widget_get_style_context (widget);

  gtk_render_background (context, cr, 0, 0, width, height);
  gtk_render_frame (context, cr, 0, 0, width, height);
}

static cairo_surface_t *
create_alpha_surface (GtkCsdWindow *window)
{
  GtkWidget *widget = GTK_WIDGET (window);
  cairo_t *cr;
  cairo_surface_t *surface;

  surface = gdk_window_create_similar_surface (gtk_widget_get_window (widget),
                                               CAIRO_CONTENT_COLOR_ALPHA,
                                               gtk_widget_get_allocated_width (widget),
                                               gtk_widget_get_allocated_height (widget));

  cr = cairo_create (surface);
  paint_background_and_frame (window, cr);
  cairo_destroy (cr);

  return surface;
}

static void
maybe_update_shape (GtkCsdWindow *window)
{
  GtkWidget *widget = GTK_WIDGET (window);
  cairo_t *cr;
  cairo_surface_t *surface;
  cairo_region_t *region;

  /* fallback to XShape only for non-composited clients */
  if (gtk_widget_is_composited (widget))
    {
      gtk_widget_shape_combine_region (widget, NULL);
      return;
    }

  surface = create_alpha_surface (window);

  region = gdk_cairo_region_create_from_surface (surface);
  gtk_widget_shape_combine_region (widget, region);

  cairo_surface_destroy (surface);
  cairo_region_destroy (region);
}



static void
gtk_csd_window_composited_changed (GtkWidget *widget)
{
  if (gtk_widget_get_realized (widget))
    maybe_update_shape (GTK_CSD_WINDOW (widget));
}

static void
gtk_csd_window_realize (GtkWidget *widget)
{
  GdkScreen *screen;
  GdkVisual *visual;

  screen = gtk_widget_get_screen (widget);
  visual = gdk_screen_get_rgba_visual (screen);
  if (visual != NULL)
    gtk_widget_set_visual (widget, visual);

  GTK_WIDGET_CLASS (gtk_csd_window_parent_class)->realize (widget);

  maybe_update_shape (GTK_CSD_WINDOW (widget));
}

static gboolean
gtk_csd_window_draw (GtkWidget *widget,
                     cairo_t   *cr)
{
  GtkCsdWindow *window = GTK_CSD_WINDOW (widget);
  cairo_surface_t *surface;

  if (gtk_widget_is_composited (widget))
    {
      cairo_save (cr);
      cairo_set_source_rgba (cr, 0, 0, 0, 0);
      cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
      cairo_paint (cr);
      cairo_restore (cr);
    }

  maybe_update_shape (window);

  cairo_push_group (cr);

  paint_background_and_frame (window, cr);

  GTK_WIDGET_CLASS (gtk_csd_window_parent_class)->draw (widget, cr);

  cairo_pop_group_to_source (cr);
  surface = create_alpha_surface (window);
  cairo_mask_surface (cr, surface, 0, 0);
  cairo_surface_destroy (surface);

  return FALSE;
}

static void
gtk_csd_window_class_init (GtkCsdWindowClass *class)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (class);
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (class);

  widget_class->realize = gtk_csd_window_realize;
  widget_class->composited_changed = gtk_csd_window_composited_changed;
  widget_class->draw = gtk_csd_window_draw;

  //g_type_class_add_private (gobject_class, sizeof (GtkCsdWindowPrivate));
}

static const gchar css_data[] =
".csd-window {\n"
"    padding: 4px;\n"
"    border-style: none;\n"
"    border-radius: 5px;\n"
"    background-color: @theme_bg_color;\n"
"    color: @theme_fg_color;\n"
"    text-shadow: 1px 1px black;\n"
"}";

static void
gtk_csd_window_init (GtkCsdWindow *window)
{
  static gboolean added = FALSE;
  GtkStyleContext *context;

  gtk_widget_set_app_paintable (GTK_WIDGET (window), TRUE);
  context = gtk_widget_get_style_context (GTK_WIDGET (window));
  gtk_style_context_add_class (context, "csd-window");

  if (!added)
    {
      GtkCssProvider *provider;
      provider = gtk_css_provider_new ();
      gtk_css_provider_load_from_data (provider,
                                       css_data, strlen (css_data),
                                       NULL);
      gtk_style_context_add_provider_for_screen (gtk_widget_get_screen (GTK_WIDGET (window)),
                                                 GTK_STYLE_PROVIDER (provider),
                                                 GTK_STYLE_PROVIDER_PRIORITY_APPLICATION);
      g_object_unref (provider);

      added = TRUE;
    }
}

GtkWidget *
gtk_csd_window_new (void)
{
  return GTK_WIDGET (g_object_new (GTK_TYPE_CSD_WINDOW,
                                   //"type", GTK_WINDOW_POPUP,
                                   NULL));
}
